'''
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.
'''

from __future__ import division

import numpy as np
from scipy.sparse import csc_matrix
from smt.surrogate_models.surrogate_model import SurrogateModel

from smt.utils.linear_solvers import get_solver
from smt.utils.caching import cached_operation

from smt.surrogate_models.rbfclib import PyRBF


class RBF(SurrogateModel):

    """
    Radial basis function interpolant with global polynomial trend.
    """

    def _initialize(self):
        super(RBF, self)._initialize()
        declare = self.options.declare
        supports = self.supports

        declare('d0', 1.0, types=(int, float, list, np.ndarray),
                desc='basis function scaling parameter in exp(-d^2 / d0^2)')
        declare('poly_degree', -1,types=int, values=(-1, 0, 1),
                desc='-1 means no global polynomial, 0 means constant, 1 means linear trend')
        declare('data_dir', values=None, types=str,
                desc='Directory for loading / saving cached data; None means do not save or load')
        declare('reg', 1e-10, types=(int, float),
                desc='Regularization coeff.')
        declare('max_print_depth', 5, types=int,
                desc='Maximum depth (level of nesting) to print operation descriptions and times')

        supports['derivatives'] = True
        supports['output_derivatives'] = True

        self.name = 'RBF'

    def _setup(self):
        options = self.options

        nx = self.training_points[None][0][0].shape[1]
        if isinstance(options['d0'], (int, float)):
            options['d0'] = [options['d0']] * nx
        options['d0'] = np.array(np.atleast_1d(options['d0']), dtype=float)

        self.printer.max_print_depth = options['max_print_depth']

        num = {}
        # number of inputs and outputs
        num['x'] = self.training_points[None][0][0].shape[1]
        num['y'] = self.training_points[None][0][1].shape[1]
        # number of radial function terms
        num['radial'] = self.training_points[None][0][0].shape[0]
        # number of polynomial terms
        if options['poly_degree'] == -1:
            num['poly'] = 0
        elif options['poly_degree'] == 0:
            num['poly'] = 1
        elif options['poly_degree'] == 1:
            num['poly'] = 1 + num['x']
        num['dof'] = num['radial'] + num['poly']

        self.num = num

        nt = self.training_points[None][0][0].shape[0]
        xt, yt = self.training_points[None][0]

        self.rbfc = PyRBF()
        self.rbfc.setup(
            num['x'], nt, num['dof'], options['poly_degree'], options['d0'], xt.flatten())

    def _new_train(self):
        num = self.num

        xt, yt = self.training_points[None][0]
        jac = np.empty(num['radial'] * num['dof'])
        self.rbfc.compute_jac(num['radial'], xt.flatten(), jac)
        jac = jac.reshape((num['radial'], num['dof']))

        mtx = np.zeros((num['dof'], num['dof']))
        mtx[:num['radial'], :] = jac
        mtx[:, :num['radial']] = jac.T
        mtx[np.arange(num['radial']), np.arange(num['radial'])] += self.options['reg']

        self.mtx = mtx

        rhs = np.zeros((num['dof'], num['y']))
        rhs[:num['radial'], :] = yt

        sol = np.zeros((num['dof'], num['y']))

        solver = get_solver('dense-lu')
        with self.printer._timed_context('Initializing linear solver'):
            solver._setup(mtx, self.printer)

        for ind_y in range(rhs.shape[1]):
            with self.printer._timed_context('Solving linear system (col. %i)' % ind_y):
                solver._solve(rhs[:, ind_y], sol[:, ind_y], ind_y=ind_y)

        self.sol = sol

    def _train(self):
        """
        Train the model
        """
        self._setup()

        tmp = self.rbfc
        self.rbfc = None

        inputs = {'self': self}
        with cached_operation(inputs, self.options['data_dir']) as outputs:
            self.rbfc = tmp

            if outputs:
                self.sol = outputs['sol']
            else:
                self._new_train()
                outputs['sol'] = self.sol

    def _predict_values(self, x):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        n = x.shape[0]
        num = self.num

        jac = np.empty(n * num['dof'])
        self.rbfc.compute_jac(n, x.flatten(), jac)
        jac = jac.reshape((n, num['dof']))

        y = jac.dot(self.sol)
        return y

    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        dy_dx : np.ndarray
            Derivative values.
        """
        n = x.shape[0]
        num = self.num

        jac = np.empty(n * num['dof'])
        self.rbfc.compute_jac_derivs(n, kx, x.flatten(), jac)
        jac = jac.reshape((n, num['dof']))

        dy_dx = jac.dot(self.sol)
        return dy_dx

    def _predict_output_derivatives(self, x):
        n = x.shape[0]
        nt = self.nt
        ny = self.training_points[None][0][1].shape[1]
        num = self.num

        dy_dstates = np.empty(n * num['dof'])
        self.rbfc.compute_jac(n, x.flatten(), dy_dstates)
        dy_dstates = dy_dstates.reshape((n, num['dof']))

        dstates_dytl = np.linalg.inv(self.mtx)

        ones = np.ones(self.nt)
        arange = np.arange(self.nt)
        dytl_dyt = csc_matrix((ones, (arange, arange)), shape=(num['dof'], self.nt))

        dy_dyt = (dytl_dyt.T.dot(dstates_dytl.T).dot(dy_dstates.T)).T
        dy_dyt = np.einsum('ij,k->ijk', dy_dyt, np.ones(ny))
        return {None: dy_dyt}



'''
Author: Dr. Mohamed A. Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
'''

import numpy as np
from sklearn.cross_decomposition.pls_ import PLSRegression as pls
from pyDOE2 import bbdesign

def standardization(X,y,copy=False):

    """
    We substract the mean from each variable. Then, we divide the values of each
    variable by its standard deviation.

    Parameters
    ----------

    X: np.ndarray [n_obs, dim]
            - The input variables.

    y: np.ndarray [n_obs, 1]
            - The output variable.

    copy: bool
            - A copy of matrices X and y will be used (copy = True).
            - Matrices X and y will be used. The matrices X and y will be
              normalized (copy = False).
            - (copy = False by default).

    Returns
    -------

    X: np.ndarray [n_obs, dim]
          The standardized input matrix.

    y: np.ndarray [n_obs, 1]
          The standardized output vector.

    X_mean: list(dim)
            The mean of each input variable.

    y_mean: list(1)
            The mean of the output variable.

    X_std:  list(dim)
            The standard deviation of each input variable.

    y_std:  list(1)
            The standard deviation of the output variable.

    """
    X_mean = np.mean(X, axis=0)
    X_std = X.std(axis=0,ddof=1)
    y_mean = np.mean(y, axis=0)
    y_std = y.std(axis=0,ddof=1)
    X_std[X_std == 0.] = 1.
    y_std[y_std == 0.] = 1.

    # center and scale X
    if copy:
        Xr = (X.copy() - X_mean) / X_std
        yr = (y.copy() - y_mean) / y_std
        return Xr, yr, X_mean, y_mean, X_std, y_std

    else:
        X = (X - X_mean) / X_std
        y = (y - y_mean) / y_std
        return X, y, X_mean, y_mean, X_std, y_std

def l1_cross_distances(X):

    """
    Computes the nonzero componentwise L1 cross-distances between the vectors
    in X.

    Parameters
    ----------

    X: np.ndarray [n_obs, dim]
            - The input variables.

    Returns
    -------

    D: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The L1 cross-distances between the vectors in X.

    ij: np.ndarray [n_obs * (n_obs - 1) / 2, 2]
            - The indices i and j of the vectors in X associated to the cross-
              distances in D.
    """

    n_samples, n_features = X.shape
    n_nonzero_cross_dist = n_samples * (n_samples - 1) // 2
    ij = np.zeros((n_nonzero_cross_dist, 2), dtype=np.int)
    D = np.zeros((n_nonzero_cross_dist, n_features))
    ll_1 = 0

    for k in range(n_samples - 1):
        ll_0 = ll_1
        ll_1 = ll_0 + n_samples - k - 1
        ij[ll_0:ll_1, 0] = k
        ij[ll_0:ll_1, 1] = np.arange(k + 1, n_samples)
        D[ll_0:ll_1] = np.abs(X[k] - X[(k + 1):n_samples])

    return D, ij.astype(np.int)


def abs_exp(theta, d):

    """
    Absolute exponential autocorrelation model.
    (Ornstein-Uhlenbeck stochastic process)::

    Parameters
    ----------
    theta : list[ncomp]
        the autocorrelation parameter(s).

    d: np.ndarray[n_obs * (n_obs - 1) / 2, n_comp]
        |d_i * coeff_pls_i| if PLS is used, |d_i| otherwise

    Returns
    -------
    r : np.ndarray[n_obs * (n_obs - 1) / 2,1]
        An array containing the values of the autocorrelation model.
    """

    r = np.zeros((d.shape[0],1))
    n_components = d.shape[1]

    # Construct/split the correlation matrix
    i,nb_limit  = 0,int(1e4)
    while True:
        if i * nb_limit > d.shape[0]:
            return r
        else:
            r[i*nb_limit:(i+1)*nb_limit,0] = np.exp(-np.sum(theta.reshape(1,
                    n_components) * d[i*nb_limit:(i+1)*nb_limit,:], axis=1))
            i+=1


def squar_exp(theta, d):

    """
    Squared exponential correlation model.

    Parameters
    ----------
    theta : list[ncomp]
        the autocorrelation parameter(s).

    d: np.ndarray[n_obs * (n_obs - 1) / 2, n_comp]
        |d_i * coeff_pls_i| if PLS is used, |d_i| otherwise

    Returns
    -------
    r: np.ndarray[n_obs * (n_obs - 1) / 2,1]
        An array containing the values of the autocorrelation model.
    """

    r = np.zeros((d.shape[0],1))
    n_components = d.shape[1]

    # Construct/split the correlation matrix
    i,nb_limit  = 0,int(1e4)

    while True:
        if i * nb_limit > d.shape[0]:
            return r
        else:
            r[i*nb_limit:(i+1)*nb_limit,0] = np.exp(-np.sum(theta.reshape(1,
                    n_components) * d[i*nb_limit:(i+1)*nb_limit,:], axis=1))
            i+=1


def ge_compute_pls(X,y,n_comp,pts,delta_x,xlimits,extra_points):

    """
    Gradient-enhanced PLS-coefficients.

    Parameters
    ----------

    X: np.ndarray [n_obs,dim]
            - - The input variables.

    y: np.ndarray [n_obs,1]
            - The output variable

    n_comp: int
            - Number of principal components used.

    pts: dict()
            - The gradient values.

    delta_x: real
            - The step used in the FOTA.

    xlimits: np.ndarray[dim, 2]
            - The upper and lower var bounds.

    extra_points: int
            - The number of extra points per each training point.

    Returns
    -------

    Coeff_pls: np.ndarray[dim, n_comp]
            - The PLS-coefficients.

    XX: np.ndarray[extra_points*nt, dim]
            - Extra points added (when extra_points > 0)

    yy: np.ndarray[extra_points*nt, 1]
            - Extra points added (when extra_points > 0)

    """
    nt,dim = X.shape
    XX = np.empty(shape = (0,dim))
    yy = np.empty(shape = (0,1))
    _pls = pls(n_comp)

    coeff_pls = np.zeros((nt,dim,n_comp))
    for i in range(nt):
        if dim >= 3:
            sign = np.roll(bbdesign(int(dim),center=1),1,axis=0)
            _X = np.zeros((sign.shape[0],dim))
            _y = np.zeros((sign.shape[0],1))
            sign = sign * delta_x*(xlimits[:,1]-xlimits[:,0])
            _X = X[i,:]+ sign
            for j in range(1,dim+1):
                sign[:,j-1] = sign[:,j-1]*pts[None][j][1][i,0]
            _y = y[i,:]+ np.sum(sign,axis=1).reshape((sign.shape[0],1))
        else:
            _X = np.zeros((9,dim))
            _y = np.zeros((9,1))
            # center
            _X[:,:] = X[i,:].copy()
            _y[0,0] = y[i,0].copy()
            # right
            _X[1,0] +=delta_x*(xlimits[0,1]-xlimits[0,0])
            _y[1,0] = _y[0,0].copy()+ pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])
            # up
            _X[2,1] +=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[2,0] = _y[0,0].copy()+ pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])
            # left
            _X[3,0] -=delta_x*(xlimits[0,1]-xlimits[0,0])
            _y[3,0] = _y[0,0].copy()- pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])
            # down
            _X[4,1] -=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[4,0] = _y[0,0].copy()-pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])
            # right up
            _X[5,0] +=delta_x*(xlimits[0,1]-xlimits[0,0])
            _X[5,1] +=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[5,0] = _y[0,0].copy()+ pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])+pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])
            # left up
            _X[6,0] -=delta_x*(xlimits[0,1]-xlimits[0,0])
            _X[6,1] +=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[6,0] = _y[0,0].copy()- pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])+pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])
            # left down
            _X[7,0] -=delta_x*(xlimits[0,1]-xlimits[0,0])
            _X[7,1] -=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[7,0] = _y[0,0].copy()- pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])-pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])
            # right down
            _X[8,0] +=delta_x*(xlimits[0,1]-xlimits[0,0])
            _X[8,1] -=delta_x*(xlimits[1,1]-xlimits[1,0])
            _y[8,0] = _y[0,0].copy()+ pts[None][1][1][i,0]*delta_x*(
                xlimits[0,1]-xlimits[0,0])-pts[None][2][1][i,0]*delta_x*(
                xlimits[1,1]-xlimits[1,0])

        _pls.fit(_X.copy(),_y.copy())
        coeff_pls[i,:,:] = _pls.x_rotations_
        #Add additional points
        if extra_points != 0:
            max_coeff = np.argsort(np.abs(coeff_pls[i,:,0]))[-extra_points:]
            for ii in max_coeff:
                XX = np.vstack((XX,X[i,:]))
                XX[-1,ii] += delta_x*(xlimits[ii,1]-xlimits[ii,0])
                yy = np.vstack((yy,y[i,0]))
                yy[-1,0] += pts[None][1+ii][1][i,0]*delta_x*(
                    xlimits[ii,1]-xlimits[ii,0])
    return np.abs(coeff_pls).mean(axis=0), XX, yy

def componentwise_distance(D,corr,dim):

    """
    Computes the nonzero componentwise cross-spatial-correlation-distance
    between the vectors in X.

    Parameters
    ----------

    D: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The cross-distances between the vectors in X depending of the correlation function.

    corr: str
            - Name of the correlation function used.
              squar_exp or abs_exp.

    dim: int
            - Number of dimension.

    Returns
    -------

    D_corr: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The componentwise cross-spatial-correlation-distance between the
              vectors in X.

    """
    # Fit the matrix iteratively: avoid some memory troubles .
    limit=int(1e4)

    D_corr = np.zeros((D.shape[0],dim))
    i,nb_limit  = 0,int(limit)

    while True:
        if i * nb_limit > D_corr.shape[0]:
            return D_corr
        else:
            if corr == 'squar_exp':
                D_corr[i*nb_limit:(i+1)*nb_limit,:] = D[i*nb_limit:(i+1)*
                                                      nb_limit,:]**2
            else:
                # abs_exp
                D_corr[i*nb_limit:(i+1)*nb_limit,:] = np.abs(D[i*nb_limit:
                                                    (i+1)*nb_limit,:])
            i+=1

def componentwise_distance_PLS(D,corr,n_comp,coeff_pls):

    """
    Computes the nonzero componentwise cross-spatial-correlation-distance
    between the vectors in X.

    Parameters
    ----------

    D: np.ndarray [n_obs * (n_obs - 1) / 2, dim]
            - The L1 cross-distances between the vectors in X.

    corr: str
            - Name of the correlation function used.
              squar_exp or abs_exp.

    n_comp: int
            - Number of principal components used.

    coeff_pls: np.ndarray [dim, n_comp]
            - The PLS-coefficients.

    Returns
    -------

    D_corr: np.ndarray [n_obs * (n_obs - 1) / 2, n_comp]
            - The componentwise cross-spatial-correlation-distance between the
              vectors in X.

    """
    # Fit the matrix iteratively: avoid some memory troubles .
    limit=int(1e4)

    D_corr = np.zeros((D.shape[0],n_comp))
    i,nb_limit  = 0,int(limit)

    while True:
        if i * nb_limit > D_corr.shape[0]:
            return D_corr
        else:
            if corr == 'squar_exp':
                D_corr[i*nb_limit:(i+1)*nb_limit,:] = np.dot(D[i*nb_limit:
                                (i+1)*nb_limit,:]** 2,coeff_pls**2)
            else:
                # abs_exp
                D_corr[i*nb_limit:(i+1)*nb_limit,:] = np.dot(np.abs(D[i*
                                nb_limit:(i+1)*nb_limit,:]),np.abs(coeff_pls))
            i+=1



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. John T. Hwang         <hwangjt@umich.edu>

This package is distributed under New BSD license.

Torsion vibration problem from:
Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design: comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition. American Society of Mechanical Engineers. June, 2016.
Wang, L., Beeson, D., Wiggs, G., and Rayasam, M. A Comparison of Metamodeling Methods Using Practical Industry Requirements. In Proceedings of the 47th AIAA/ASME/ASCE/AHS/ASC structures, structural dynamics, and materials conference, Newport, RI, pp. AIAA 2006-1811.
"""
from __future__ import division
import numpy as np
from scipy.misc import derivative

from smt.problems.problem import Problem

class TorsionVibration(Problem):

    def _initialize(self):
        self.options.declare('name', 'TorsionVibration', types=str)
        self.options.declare('use_FD', False, types=bool)
        self.options['ndim'] = 15

    def _setup(self):
        assert self.options['ndim'] == 15, 'ndim must be 15'

        self.xlimits[:, 0] = \
            [1.8,9,10530000,7.2,3510000,10.8,1.6425,10.8,5580000,2.025,2.7,0.252,12.6,3.6,0.09]
        self.xlimits[:, 1] = \
            [2.2,11,12870000,8.8,4290000,13.2,2.0075,13.2,6820000,2.475,3.3,0.308,15.4,4.4,0.11]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        y = np.zeros((ne, 1), complex)

        def partial_derivative(function, var=0, point=[]):
            args = point[:]
            def wraps(x):
                args[var] = x
                return func(*args)
            return derivative(wraps, point[var], dx = 1e-6)

        def func(x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14):
            K1 = np.pi*x2*x0/(32*x1)
            K2 = np.pi*x8*x6/(32*x7)
            K3 = np.pi*x4*x9/(32*x3)
            M1 = x11*np.pi*x10*x5/(4*9.80665)
            M2 = x14*np.pi*x13*x12/(4*9.80665)
            J1 = 0.5*M1*(x5/2)**2
            J2 = 0.5*M2*(x12/2)**2
            a = 1
            b = -((K1+K2)/J1+(K2+K3)/J2)
            c = (K1*K2+K2*K3+K3*K1)/(J1*J2)
            return np.sqrt((-b-np.sqrt(b**2-4*a*c))/(2*a))/(2*np.pi)

        for i in range(ne):
            x0 = x[i, 0]
            x1 = x[i, 1]
            x2 = x[i, 2]
            x3 = x[i, 3]
            x4 = x[i, 4]
            x5 = x[i, 5]
            x6 = x[i, 6]
            x7 = x[i, 7]
            x8 = x[i, 8]
            x9 = x[i, 9]
            x10 = x[i, 10]
            x11 = x[i, 11]
            x12 = x[i, 12]
            x13 = x[i, 13]
            x14 = x[i, 14]
            if kx is None:
                y[i, 0] = func(x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14)
            else:
                point = [x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13, x14]
                if self.options['use_FD']:
                    point = np.real(np.array(point))
                    y[i, 0] = partial_derivative(func, var=kx, point=point)
                else:
                    ch = 1e-20
                    point[kx] += complex(0, ch)
                    y[i, 0] = np.imag(func(*point)) / ch
                    point[kx] -= complex(0, ch)

        return y



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. Nathalie.bartoli      <nathalie@onera.fr>

This package is distributed under New BSD license.

TO DO:
- define outputs['sol'] = self.sol
"""

from __future__ import division

import numpy as np
import scipy
from smt.surrogate_models.surrogate_model import SurrogateModel
from smt.utils.caching import cached_operation


class QP(SurrogateModel):

    """
    Square polynomial approach
    """

    def _initialize(self):
        super(QP, self)._initialize()
        declare = self.options.declare
        supports = self.supports

        declare('data_dir', values=None, types=str,
                desc='Directory for loading / saving cached data; None means do not save or load')

        self.name = 'QP'
        supports['derivatives'] = True

    ############################################################################
    # Model functions
    ############################################################################

    def _new_train(self):

        """
        Train the model
        """

        if 0 in self.training_points[None]:
            x = self.training_points[None][0][0]
            y = self.training_points[None][0][1]

        if x.shape[0] < (self.nx+1)*(self.nx+2)/2.:
            raise Exception("Number of training points should be greater or equal to %d."
                            % ((self.nx+1)*(self.nx+2)/2.))

        X = self._response_surface(x)
        self.coef = np.dot(np.linalg.inv(np.dot(X.T,X)),(np.dot(X.T,y)))

    def _train(self):
        """
        Train the model
        """
        inputs = {'self': self}
        with cached_operation(inputs, self.options['data_dir']) as outputs:
            if outputs:
                self.sol = outputs['sol']
            else:
                self._new_train()
                #outputs['sol'] = self.sol

    def _response_surface(self,x):
        """
        Build the response surface of degree 2
        argument
        -----------
        x : np.ndarray [nt, dim]
            Training points
        Returns
        -------
        M : np.ndarray
            Matrix of the surface
        """
        dim = self.nx
        n = x.shape[0]
        n_app = int(scipy.special.binom(dim+2, dim))
        M = np.zeros((n_app,n))
        x = x.T
        M[0,:] = np.ones((1,n))
        for i in range(1,dim+1):
            M[i,:] = x[i-1,:]
        for i in range(dim+1,2*dim+1):
            M[i,:]=x[i-(dim+1),:]**2
        for i in range(dim-1):
            for j in range(i+1,dim):
                k = int(2*dim+2+(i)*dim-((i+1)*(i))/2+(j-(i+2)))
                M[k,:] = x[i,:]*x[j,:]

        return M.T


    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray
            Derivative values.
        """
        dim = self.nx

        linear_coef = self.coef[1+kx,0]
        quad_coef = 2 * self.coef[1+dim+kx,0] * x[:,kx]
        cross_coef = 0
        for i in range(dim):
            if i > kx:
                k = int(2*dim+2+(kx)*dim-((kx+1)*(kx))/2+(i-(kx+2)))
                cross_coef += self.coef[k,0]* x[:,i]
            elif i < kx:
                k = int(2*dim+2+(i)*dim-((i+1)*(i))/2+(kx-(i+2)))
                cross_coef += self.coef[k,0]* x[:,i]

        y = (linear_coef+quad_coef+cross_coef).reshape((x.shape[0],1))
        return y

    def _predict_values(self, x):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """

        M=self._response_surface(x)
        y = np.dot(M,self.coef).T

        return y



# -*- coding: utf-8 -*-
"""
Created on Fri May 04 10:26:49 2018

@author: Mostafa Meliani <melimostafa@gmail.com>
Multi-Fidelity co-Kriging: recursive formulation with autoregressive model of 
order 1 (AR1)
"""

from __future__ import division
import numpy as np
from smt.surrogate_models.krg_based import KrgBased
from types import FunctionType
from smt.utils.kriging_utils import l1_cross_distances, componentwise_distance,\
    standardization
from scipy.linalg import solve_triangular
from scipy import linalg
from sklearn.metrics.pairwise import manhattan_distances
import copy 
from copy import deepcopy
from sys import exit

"""
The MFK class.
"""


class MFK(KrgBased):

    """
    - MFK
    """
    def _initialize(self):
        super(MFK, self)._initialize()
        declare = self.options.declare
        
        declare('rho_regr', 'constant',types=FunctionType,\
                values=('constant', 'linear', 'quadratic'), desc='regr. term')
        declare('theta0', None, types=(list, np.ndarray), \
                desc='Initial hyperparameters')
        declare('optim_var', False, types = bool, \
                values = (True, False), \
                desc ='Turning this option to True, forces variance to zero at HF samples ')
        declare('eval_noise', False, types = bool, \
                values = (True, False), desc ='noise evaluation flag')
        declare('noise0', 1e-6, types = float, \
                desc ='Initial noise hyperparameter')
        self.name = 'MFK'
    

   

    def _check_list_structure(self, X, y):
        """
        checks if the data structure is compatible with MFK.
        sets class attributes such as (number of levels of Fidelity, training points in each level, ...)
        
        Arguments :
        X : list of arrays, each array corresponds to a fidelity level. starts from lowest to highest
        y : same as X 
        """
        

        if type(X) is not list:
            nlevel = 1
            X = [X]
        else:
            nlevel = len(X)


        if type(y) is not list:
            y = [y]

        if len(X) != len(y):
            raise ValueError("X and y must have the same length.")


        n_samples = np.zeros(nlevel, dtype = int)
        n_features = np.zeros(nlevel, dtype = int)
        n_samples_y = np.zeros(nlevel, dtype = int)
        for i in range(nlevel):
            n_samples[i], n_features[i] = X[i].shape
            if i>1 and n_features[i] != n_features[i-1]:
                raise ValueError("All X must have the same number of columns.")
            y[i] = np.asarray(y[i]).ravel()[:, np.newaxis]
            n_samples_y[i] = y[i].shape[0]
            if n_samples[i] != n_samples_y[i]:
                raise ValueError("X and y must have the same number of rows.")


        self.nx = n_features[0]
        self.nt_all = n_samples
        self.nlvl = nlevel
        self.ny = y[0].shape[1]
        self.X = X[:]
        self.y = y[:]
        
    
    def _new_train(self):
        """
        Overrides KrgBased implementation
        Trains the Multi-Fidelity model
        """
        
        xt =[]
        yt = []
        i=0
        while(self.training_points.get(i, None) is not None):
            xt.append(self.training_points[i][0][0])
            yt.append(self.training_points[i][0][1])
            i = i+1
        xt.append(self.training_points[None][0][0])
        yt.append(self.training_points[None][0][1])
            
        
        self._check_list_structure(xt, yt)
        self._check_param()
        X = self.X
        y = self.y
        
        _, _, self.X_mean, self.y_mean, self.X_std, \
            self.y_std = standardization(np.concatenate(xt,axis=0), np.concatenate(yt,axis=0))
        
        
        nlevel = self.nlvl
        n_samples = self.nt_all

        # initialize lists
        self.noise = nlevel*[0]
        self.D_all = nlevel*[0]
        self.F_all = nlevel*[0]
        self.p_all = nlevel*[0]
        self.q_all = nlevel*[0]
        self.optimal_rlf_value = nlevel*[0]
        self.optimal_par = nlevel*[{}]
        self.optimal_theta = nlevel*[0]
        self.X_norma_all = [(x-self.X_mean)/self.X_std for x in X] 
        self.y_norma_all = [(f-self.y_mean)/self.y_std for f in y] 

        for lvl in range(nlevel):
            self.X_norma = self.X_norma_all[lvl]
            self.y_norma = self.y_norma_all[lvl]
            # Calculate matrix of distances D between samples
            self.D_all[lvl] = l1_cross_distances(self.X_norma)
            

            # Regression matrix and parameters
            self.F_all[lvl] = self.options['poly'](self.X_norma)
            self.p_all[lvl] = self.F_all[lvl].shape[1]

            # Concatenate the autoregressive part for levels > 0
            if lvl > 0:
                F_rho = self.options['rho_regr'](self.X_norma)
                self.q_all[lvl] = F_rho.shape[1]
                self.F_all[lvl] = np.hstack((F_rho*np.dot(self._predict_intermediate_values(self.X_norma, lvl, descale=False),
                                              np.ones((1,self.q_all[lvl]))), self.F_all[lvl]))
            else:
                self.q_all[lvl] = 0

            n_samples_F_i = self.F_all[lvl].shape[0]

            if n_samples_F_i != n_samples[lvl]:
                raise Exception("Number of rows in F and X do not match. Most "
                                "likely something is going wrong with the "
                                "regression model.")

            if int(self.p_all[lvl] + self.q_all[lvl]) >= n_samples_F_i:
                raise Exception(("Ordinary least squares problem is undetermined "
                                 "n_samples=%d must be greater than the regression"
                                 " model size p+q=%d.")
                                 % (n_samples[i], self.p_all[lvl]+self.q_all[lvl]))

       
            # Determine Gaussian Process model parameters
            self.F = self.F_all[lvl]
            D, self.ij = self.D_all[lvl]
            self._lvl = lvl
            self.nt = self.nt_all[lvl]
            self.q = self.q_all[lvl]
            self.p = self.p_all[lvl]
            self.optimal_rlf_value[lvl], self.optimal_par[lvl], self.optimal_theta[lvl] = \
                self._optimize_hyperparam(D)
            if self.options['eval_noise']:
                tmp_list = self.optimal_theta[lvl]
                self.optimal_theta[lvl] = tmp_list[:-1]
                self.noise[lvl] = tmp_list[-1]
            del self.y_norma, self.D
        
        if self.options['eval_noise'] and self.options['optim_var']:
            for lvl in range(self.nlvl-1):
                self.set_training_values(X[lvl], self._predict_intermediate_values(X[lvl], lvl+1), name = lvl)
            self.set_training_values(X[-1], self._predict_intermediate_values(X[-1], self.nlvl))   
            self.options['eval_noise'] = False
            self._new_train()
            
    def _componentwise_distance(self,dx,opt=0):
        d = componentwise_distance(dx,self.options['corr'].__name__,
                                   self.nx)
        return d
    
    def _predict_intermediate_values(self, X, lvl, descale = True):
        """
        Evaluates the model at a set of points.
        Used for training the model at level lvl.
        Allows to relax the order problem.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        lvl : level at which the prediction is made

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        n_eval, _ = X.shape
#        if n_features_X != self.n_features:
#            raise ValueError("Design must be an array of n_features columns.")

        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, lvl))
#        if self.normalize:
        if descale :
            X = (X - self.X_mean) / self.X_std
##                X = (X - self.X_mean[0]) / self.X_std[0]
        f = self.options['poly'](X)
        f0 = self.options['poly'](X)
        dx = manhattan_distances(X, Y=self.X_norma_all[0], sum_over_features=False)
        d = self._componentwise_distance(dx)
        # Get regression function and correlation
        F = self.F_all[0]
        C = self.optimal_par[0]['C']

        beta = self.optimal_par[0]['beta']
        Ft = solve_triangular(C, F, lower=True)
        yt = solve_triangular(C, self.y_norma_all[0], lower=True)
        r_ = self.options['corr'](self.optimal_theta[0], d).reshape(n_eval, self.nt_all[0])
        gamma = self.optimal_par[0]['gamma']
        

        # Scaled predictor
        mu[:,0]= (np.dot(f, beta) + np.dot(r_,gamma)).ravel()

        # Calculate recursively kriging mean and variance at level i
        for i in range(1,lvl):
            F = self.F_all[i]
            C = self.optimal_par[i]['C']
            g = self.options['rho_regr'](X)
            dx = manhattan_distances(X, Y=self.X_norma_all[i], sum_over_features=False)
            d = self._componentwise_distance(dx)
            r_ = self.options['corr'](self.optimal_theta[i], d).reshape(n_eval, self.nt_all[i])
            f = np.vstack((g.T*mu[:,i-1], f0.T))
            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y_norma_all[i], lower=True)
            beta = self.optimal_par[i]['beta']
            gamma = self.optimal_par[i]['gamma']
            # scaled predictor
            mu[:,i] = (np.dot(f.T, beta) \
                       + np.dot(r_, gamma)).ravel()

        # scaled predictor
        if descale :
            for i in range(lvl):# Predictor
                mu[:,i] = self.y_mean + self.y_std * mu[:,i]
           
        return mu[:,-1].reshape((n_eval,1))
        
        
        

    def _predict_values(self, X):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        
        return self._predict_intermediate_values( X, self.nlvl)


    def _predict_variances(self, X):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        return self.predict_variances_all_levels(X)[0][:,-1]
        
    def predict_variances_all_levels(self, X):
        """
        Evaluates the model at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values

        Returns
        -------
        y : np.ndarray
            Evaluation point output variable values
        """
        # Initialization X = atleast_2d(X)
        nlevel = self.nlvl
        sigma2_rhos =[]
        n_eval, n_features_X = X.shape
#        if n_features_X != self.n_features:
#            raise ValueError("Design must be an array of n_features columns.")
        X = (X - self.X_mean) / self.X_std
        # Calculate kriging mean and variance at level 0
        mu = np.zeros((n_eval, nlevel))
#        if self.normalize:
        f = self.options['poly'](X)
        f0 = self.options['poly'](X)
        dx = manhattan_distances(X, Y=self.X_norma_all[0], sum_over_features=False)
        d = self._componentwise_distance(dx)
        # Get regression function and correlation
        F = self.F_all[0]
        C = self.optimal_par[0]['C']

        beta = self.optimal_par[0]['beta']
        Ft = solve_triangular(C, F, lower=True)
        yt = solve_triangular(C, self.y_norma_all[0], lower=True)
        r_ = self.options['corr'](self.optimal_theta[0], d).reshape(n_eval, self.nt_all[0])
        gamma = self.optimal_par[0]['gamma']

        # Scaled predictor
        mu[:,0]= (np.dot(f, beta) + np.dot(r_,gamma)).ravel()

    
        self.sigma2_rho = nlevel*[None]
        MSE = np.zeros((n_eval,nlevel))
        r_t = solve_triangular(C, r_.T, lower=True)
        G = self.optimal_par[0]['G']

        u_ = solve_triangular(G.T, f.T - np.dot(Ft.T, r_t), lower=True)
        MSE[:,0] = self.optimal_par[0]['sigma2'] * (1 + self.noise[0]\
                            - (r_t**2).sum(axis=0) + (u_**2).sum(axis=0))
        

        # Calculate recursively kriging variance at level i
        for i in range(1,nlevel):
            F = self.F_all[i]
            C = self.optimal_par[i]['C']
            g = self.options['rho_regr'](X)
            dx = manhattan_distances(X, Y=self.X_norma_all[i], sum_over_features=False)
            d = self._componentwise_distance(dx)
            r_ = self.options['corr'](self.optimal_theta[i], d).reshape(n_eval, self.nt_all[i])
            f = np.vstack((g.T*mu[:,i-1], f0.T))

            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y_norma_all[i], lower=True)
            r_t = solve_triangular(C,r_.T, lower=True)
            G = self.optimal_par[i]['G']
            beta = self.optimal_par[i]['beta']

            # scaled predictor
            sigma2 = self.optimal_par[i]['sigma2'] 
            q = self.q_all[i]
            p = self.p_all[i]
            Q_ = (np.dot((yt-np.dot(Ft,beta)).T, yt-np.dot(Ft,beta)))[0,0]
            u_ = solve_triangular(G.T, f - np.dot(Ft.T, r_t), lower=True)
            sigma2_rho = np.dot(g, \
                sigma2*linalg.inv(np.dot(G.T,G))[:q,:q] \
                    + np.dot(beta[:q], beta[:q].T))
            sigma2_rho = (sigma2_rho * g).sum(axis=1)
            sigma2_rhos.append(sigma2_rho)
            
            MSE[:,i] = sigma2_rho * MSE[:,i-1] \
                            + Q_/(2*(self.nt_all[i]-p-q)) \
                            * (1 + self.noise[i] - (r_t**2).sum(axis=0)) \
                            + sigma2 * (u_**2).sum(axis=0)

                

        # scaled predictor
        for i in range(nlevel):# Predictor
            MSE[:,i] = self.y_std**2 * MSE[:,i]
      
        return MSE, sigma2_rhos
    
    
    def _predict_derivatives(self, x, kx):
        """
        Evaluates the derivatives at a set of points.

        Arguments
        ---------
        x : np.ndarray [n_evals, dim]
            Evaluation point input variable values
        kx : int
            The 0-based index of the input variable with respect to which derivatives are desired.

        Returns
        -------
        y : np.ndarray*self.y_std/self.X_std[kx])
            Derivative values.
        """
        
        
        lvl =self.nlvl
        # Initialization
        
        n_eval, n_features_x = x.shape
        x = (x - self.X_mean) / self.X_std
        
        dy_dx = np.zeros((n_eval, lvl))
        
        if self.options['corr'].__name__ != 'squar_exp':
            raise ValueError(
            'The derivative is only available for square exponential kernel')
        if self.options['poly'].__name__ == 'constant':
            df = np.zeros([n_eval,1])
        elif self.options['poly'].__name__ == 'linear':
            df = np.zeros((n_eval, self.nx + 1))
            df[:,1:] = 1
        else:
            raise ValueError(
                'The derivative is only available for ordinary kriging or '+
                'universal kriging using a linear trend')
        df0 = copy.deepcopy(df)
        if self.options['rho_regr'].__name__ != 'constant' :
            raise ValueError(
                'The derivative is only available for regression rho constant')
        # Get pairwise componentwise L1-distances to the input training set
        dx = manhattan_distances(x, Y=self.X_norma_all[0], sum_over_features=
                                 False)
        d = self._componentwise_distance(dx)
        # Compute the correlation function
        r_ = self.options['corr'](self.optimal_theta[0], d).reshape(n_eval,self.nt_all[0])

        # Beta and gamma = R^-1(y-FBeta)
        beta = self.optimal_par[0]['beta']
        gamma = self.optimal_par[0]['gamma']

        df_dx = np.dot(df, beta)
        d_dx=x[:,kx].reshape((n_eval,1))-self.X_norma_all[0][:,kx].reshape((1,self.nt_all[0]))
        theta = self.optimal_theta[0]

        dy_dx[:,0] = np.ravel((df_dx-2*theta[kx]*np.dot(d_dx*r_,gamma)))



        # Calculate recursively derivative at level i
        for i in range(1,lvl):
            F = self.F_all[i]
            C = self.optimal_par[i]['C']
            g = self.options['rho_regr'](x)
            dx = manhattan_distances(x, Y=self.X_norma_all[i], sum_over_features=False)
            d = self._componentwise_distance(dx)
            r_ = self.options['corr'](self.optimal_theta[i], d).reshape(n_eval, self.nt_all[i])
            df = np.vstack((g.T*dy_dx[:,i-1], df0.T))

            Ft = solve_triangular(C, F, lower=True)
            yt = solve_triangular(C, self.y_norma_all[i], lower=True)
            beta = self.optimal_par[i]['beta']
            gamma = self.optimal_par[i]['gamma']
            
            df_dx = np.dot(df.T, beta)
            d_dx=x[:,kx].reshape((n_eval,1))-self.X_norma_all[i][:,kx].reshape((1,self.nt_all[i]))
            theta = self.optimal_theta[i]
            # scaled predictor
            dy_dx[:,i] = np.ravel(df_dx-2*theta[kx]*np.dot(d_dx*r_,gamma))
       
        
           
        return dy_dx[:,-1]*self.y_std/self.X_std[kx]
        



"""
Author: Dr. John T. Hwang <hwangjt@umich.edu>

This package is distributed under New BSD license.

Full-factorial sampling.
"""
from __future__ import division
import numpy as np
from six.moves import range

from smt.sampling_methods.sampling_method import SamplingMethod

class FullFactorial(SamplingMethod):

    def _initialize(self):
        self.options.declare('weights', values=[None], types=[list, np.ndarray])
        self.options.declare('clip', types=bool)

    def _compute(self, nt):
        """
        Compute the requested number of sampling points.

        Arguments
        ---------
        nt : int
            Number of points requested.

        Returns
        -------
        ndarray[nt, nx]
            The sampling locations in the input space.
        """
        xlimits = self.options['xlimits']
        nx = xlimits.shape[0]

        if self.options['weights'] is None:
            weights = np.ones(nx) / nx
        else:
            weights = np.atleast_1d(self.options['weights'])
            weights /= numpy.sum(weights)

        num_list = np.ones(nx, int)
        while np.prod(num_list) < nt:
            ind = np.argmax(weights - num_list / np.sum(num_list))
            num_list[ind] += 1

        lins_list = [np.linspace(0., 1., num_list[kx]) for kx in range(nx)]
        x_list = np.meshgrid(*lins_list, indexing='ij')

        if self.options['clip']:
            nt = np.prod(num_list)

        x = np.zeros((nt, nx))
        for kx in range(nx):
            x[:, kx] = x_list[kx].reshape(np.prod(num_list))[:nt]

        return x



"""
Author: Dr. Mohamed Amine Bouhlel <mbouhlel@umich.edu>
        Dr. John T. Hwang         <hwangjt@umich.edu>

This package is distributed under New BSD license.

Aircraft wing weight problem from:
Liu, H., Xu, S., & Wang, X. Sampling strategies and metamodeling techniques for engineering design: comparison and application. In ASME Turbo Expo 2016: Turbomachinery Technical Conference and Exposition. American Society of Mechanical Engineers. June, 2016.
Forrester, A., Sobester, A., and Keane, A., 2008,
Engineering Design Via Surrogate Modelling: A Practical Guide, John Wiley & Sons, United Kingdom.
"""

from __future__ import division
import numpy as np
from scipy.misc import derivative

from smt.problems.problem import Problem

class WingWeight(Problem):

    def _initialize(self):
        self.options.declare('name', 'WingWeight', types=str)
        self.options.declare('use_FD', False, types=bool)
        self.options['ndim'] = 10

    def _setup(self):
        assert self.options['ndim'] == 10, 'ndim must be 10'

        self.xlimits[:, 0] = [150,220,6,-10,16,0.5,0.08,2.5,1700,0.025]
        self.xlimits[:, 1] = [200,300,10,10,45,1,0.18,6,2500,0.08]

    def _evaluate(self, x, kx):
        """
        Arguments
        ---------
        x : ndarray[ne, nx]
            Evaluation points.
        kx : int or None
            Index of derivative (0-based) to return values with respect to.
            None means return function value rather than derivative.

        Returns
        -------
        ndarray[ne, 1]
            Functions values if kx=None or derivative values if kx is an int.
        """
        ne, nx = x.shape

        y = np.zeros((ne, 1), complex)

        def deg2rad(deg):
            rad = deg / 180. * np.pi
            return rad

        def partial_derivative(function, var=0, point=[]):
            args = point[:]
            def wraps(x):
                args[var] = x
                return func(*args)
            return derivative(wraps, point[var], dx = 1e-6)

        def func(x0,x1,x2,x3,x4,x5,x6,x7,x8,x9):
            return 0.036*x0**0.758*x1**0.0035*(x2/np.cos(deg2rad(x3))**2) \
                *x4**0.006*x5**0.04*(100*x6/np.cos(deg2rad(x3)))**(-0.3)*(x7*x8)**0.49+x0*x9

        for i in range(ne):
            x0 = x[i,0]
            x1 = x[i,1]
            x2 = x[i,2]
            x3 = x[i,3]
            x4 = x[i,4]
            x5 = x[i,5]
            x6 = x[i,6]
            x7 = x[i,7]
            x8 = x[i,8]
            x9 = x[i,9]
            if kx is None:
                y[i,0] = func(x0,x1,x2,x3,x4,x5,x6,x7,x8,x9)
            else:
                point = [x0,x1,x2,x3,x4,x5,x6,x7,x8,x9]
                if self.options['use_FD']:
                    point = np.real(np.array(point))
                    y[i, 0] = partial_derivative(func, var=kx, point=point)
                else:
                    ch = 1e-20
                    point[kx] += complex(0, ch)
                    y[i, 0] = np.imag(func(*point)) / ch
                    point[kx] -= complex(0, ch)

        return y



'''
Author: Dr. Mohamed A. Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
'''

from __future__ import division
import warnings
import numpy as np
from smt.surrogate_models.krg_based import KrgBased
from smt.utils.kriging_utils import componentwise_distance_PLS, ge_compute_pls

"""
The KPLS class.
"""

class GEKPLS(KrgBased):

    """
    - GEKPLS
    """
    def _initialize(self):
        super(GEKPLS, self)._initialize()
        declare = self.options.declare
        declare('xlimits', types=np.ndarray,
                desc='Lower/upper bounds in each dimension - ndarray [nx, 2]')
        declare('n_comp', 1, types=int, desc='Number of principal components')
        declare('theta0', [1e-2], types=(list, np.ndarray), desc='Initial hyperparameters')
        declare('delta_x', 1e-4, types=(int, float), desc='Step used in the FOTA')
        declare('extra_points', 0, types=int, desc='Number of extra points per training point')
        self.supports['training_derivatives'] = True

        self.name = 'GEKPLS'

    def _compute_pls(self,X,y):
        if 0 in self.training_points[None]:
            self.coeff_pls, XX, yy = ge_compute_pls(X.copy(),y.copy(),self.options['n_comp'],
                self.training_points,self.options['delta_x'],self.options['xlimits'],
                self.options['extra_points'])
            if self.options['extra_points'] != 0:
                self.nt *= (self.options['extra_points']+1)
                X = np.vstack((X,XX))
                y = np.vstack((y,yy))

        return X,y

    def _componentwise_distance(self,dx,opt=0):

        d = componentwise_distance_PLS(dx,self.options['corr'].__name__,
                                       self.options['n_comp'],self.coeff_pls)
        return d



'''
Author: Dr. Mohamed A. Bouhlel <mbouhlel@umich.edu>

This package is distributed under New BSD license.
'''

from __future__ import division
import warnings
import numpy as np

from smt.surrogate_models.krg_based import KrgBased
from smt.utils.kriging_utils import componentwise_distance

"""
The kriging class.
"""

class KRG(KrgBased):

    def _initialize(self):
        super(KRG, self)._initialize()
        declare = self.options.declare
        declare('theta0', None, types=(list, np.ndarray), desc='Initial hyperparameters')
        self.name = 'Kriging'

    def _componentwise_distance(self,dx,opt=0):
        d = componentwise_distance(dx,self.options['corr'].__name__,
                                   self.nx)
        return d
